# -*- coding: utf-8 -*-
"""
Summary
-------

Public (front-end) functions to calculate Spectrum with HITRAN / CDSD databanks.
Uses the SpectrumFactory classe from `factory.py`, Spectrum from `spectrum.py`
and line survey from `line_survey.py`

Routine Listing
---------------

:func:`~radis.lbl.calc.calc_spectrum`

-------------------------------------------------------------------------------

"""


from copy import deepcopy
from os.path import exists

from radis.lbl.factory import SpectrumFactory
from radis.misc.basics import all_in
from radis.phys.convert import nm2cm


# %%
def calc_spectrum(
    wavenum_min=None,
    wavenum_max=None,
    wavelength_min=None,
    wavelength_max=None,
    Tgas=None,
    Tvib=None,
    Trot=None,
    pressure=1.01325,
    molecule=None,
    isotope="all",
    mole_fraction=1,
    path_length=1,
    medium="air",
    databank="fetch",
    wstep=0.01,
    broadening_max_width=10,
    optimization="min-RMS",
    overpopulation=None,
    name=None,
    use_cached=True,
    verbose=True,
    mode="cpu",
    **kwargs
):
    """Multipurpose function to calculate :class:`~radis.spectrum.spectrum.Spectrum`
        under equilibrium (using either CPU or GPU), or non-equilibrium, with or without overpopulation.
        It's a wrapper to :class:`~radis.lbl.factory.SpectrumFactory` class.
        For advanced used, please refer to the aforementionned class.
    ​
        Parameters
        ----------
        wavenum_min: float [cm-1]
            minimum wavenumber to be processed in cm^-1
        wavenum_max: float [cm-1]
            maximum wavenumber to be processed in cm^-1
    ​
        wavelength_min: float [nm]
            minimum wavelength to be processed in nm. Wavelength in ``'air'`` or
            ``'vacuum'`` depending of the value of the parameter ``'medium='``
    ​
        wavelength_max: float [nm]
            maximum wavelength to be processed in nm. Wavelength in ``'air'`` or
            ``'vacuum'`` depending of the value of the parameter ``'medium='``
    ​
        Tgas: float [K]
            Gas temperature. If non equilibrium, is used for Ttranslational.
            Default ``300`` K
    ​
        Tvib: float [K]
            Vibrational temperature. If ``None``, equilibrium calculation is run with Tgas
    ​
        Trot: float [K]
            Rotational temperature. If ``None``, equilibrium calculation is run with Tgas
    ​
        pressure: float [bar]
            partial pressure of gas in bar. Default ``1.01325`` (1 atm)
    ​
        molecule: int, str, list or ``None``
            molecule id (HITRAN format) or name. For multiple molecules, use a list.
            The `isotope`, `mole_fraction`, `databank` and `overpopulation` parameters must then
            be dictionaries.
            If ``None``, the molecule can be infered
            from the database files being loaded. See the list of supported molecules
            in :py:data:`~radis.db.MOLECULES_LIST_EQUILIBRIUM`
            and :py:data:`~radis.db.MOLECULES_LIST_NONEQUILIBRIUM`.
            Default ``None``.
    ​
        isotope: int, list, str of the form ``'1,2'``, or ``'all'``, or dict
            isotope id (sorted by relative density: (eg: 1: CO2-626, 2: CO2-636 for CO2).
            See [HITRAN-2016]_ documentation for isotope list for all species. If ``'all'``,
            all isotopes in database are used (this may result in larger computation
            times!). Default ``'all'``.

            For multiple molecules, use a dictionary with molecule names as keys.

            Example::

                mole_fraction={'CO2':0.8 ,  'CO':0.2 }
    ​
        mole_fraction: float or dict
            database species mole fraction. Default ``1``.

            For multiple molecules, use a dictionary with molecule names as keys.

            Example::

                mole_fraction={'CO2': 0.8, 'CO':0.2}
    ​
        path_length: float [cm]
            slab size. Default ``1``.
    ​
        databank: str or dict
            can be either:
    ​
            - ``'fetch'``, to fetch automatically from [HITRAN-2016]_ through astroquery.

            .. warning::

                [HITRAN-2016]_ is valid for low temperatures (typically < 700 K). For higher
                temperatures you may need [HITEMP-2010]_

            - the name of a a valid database file, in which case the format is inferred.
              For instance, ``'.par'`` is recognized as ``hitran/hitemp`` format.
              Accepts wildcards ``'*'`` to select multiple files.

            - the name of a spectral database registered in your ``~/.radis``
              configuration file. This allows to use multiple database files.
              See :ref:`Configuration file <label_lbl_config_file>`.
    ​
            Default ``'fetch'``. See :class:`~radis.lbl.loader.DatabankLoader` for more
            information on line databases, and :data:`~radis.misc.config.DBFORMAT` for
            your ``~/.radis`` file format


            For multiple molecules, use a dictionary with molecule names as keys::
    ​
                databank='fetch'     # automatic download
                databank='PATH/TO/05_HITEMP2019.par'    # path to a file
                databank='*CO2*.par' #to get all the files that have CO2 in their names (case insensitive)
                databank='HITEMP-2019-CO'   # user-defined database in Configuration file
                databank = {'CO2' : 'PATH/TO/05_HITEMP2019.par', 'CO' : 'fetch'}  # for multiple molecules
    ​
        medium: ``'air'``, ``'vacuum'``
            propagating medium when giving inputs with ``'wavenum_min'``, ``'wavenum_max'``.
            Does not change anything when giving inputs in wavenumber. Default ``'air'``
    ​
        wstep: float (cm-1)
            Spacing of calculated spectrum. Default ``0.01 cm-1``
    ​
        broadening_max_width: float (cm-1)
            Full width over which to compute the broadening. Large values will create
            a huge performance drop (scales as ~broadening_width^2 without DLM)
            The calculated spectral range is increased (by broadening_max_width/2
            on each side) to take into account overlaps from out-of-range lines.
            Default ``10`` cm-1.
    ​
        Other Parameters
        ----------------
    ​
        optimization : ``"simple"``, ``"min-RMS"``, ``None``
            If either ``"simple"`` or ``"min-RMS"`` DLM optimization for lineshape calculation is used:
            - ``"min-RMS"`` : weights optimized by analytical minimization of the RMS-error (See: [DLM_article]_)
            - ``"simple"`` : weights equal to their relative position in the grid

            If using the DLM optimization, broadening method is automatically set to ``'fft'``.
            If ``None``, no lineshape interpolation is performed and the lineshape of all lines is calculated.

            Refer to [DLM_article]_ for more explanation on the DLM method for lineshape interpolation.

            Default ``"min-RMS"``
    ​
        overpopulation: dict
            dictionary of overpopulation compared to the given vibrational temperature.
            Default ``None``.

            Example::

                overpopulation = {'CO2' : {'(00`0`0)->(00`0`1)': 2.5,
                                           '(00`0`1)->(00`0`2)': 1,
                                           '(01`1`0)->(01`1`1)': 1,
                                           '(01`1`1)->(01`1`2)': 1

                                            }
                                 }
    ​
        slit: float, str, or ``None``
            if float, FWHM of a triangular slit function. If str, path to an
            experimental slit function. If None, no slit is applied. Default ``None``.
    ​
        plot: str
            any parameter such as 'radiance' (if slit is given), 'radiance_noslit',
            'absorbance', etc...   Default ``None``
    ​
        name: str
            name of the case. If None, a unique ID is generated. Default ``None``
    ​
        use_cached: boolean
            use cached files for line database and energy database. Default ``True``
    ​
        verbose: boolean, or int
            If ``False``, stays quiet. If ``True``, tells what is going on.
            If ``>=2``, gives more detailed messages (for instance, details of
            calculation times). Default ``True``.
    ​
        **kwargs: other inputs forwarded to SpectrumFactory
            For instance: ``warnings``.
            See :class:`~radis.lbl.factory.SpectrumFactory` documentation for more
            details on input.
            For instance:
    ​
        pseudo_continuum_threshold: float
            if not 0, first calculate a rough approximation of the spectrum, then
            moves all lines whose linestrength intensity is less than this threshold
            of the maximum in a semi-continuum. Values above 0.01 can yield significant
            errors, mostly in highly populated areas. 80% of the lines can typically
            be moved in a continuum, resulting in 5 times faster spectra. If 0,
            no semi-continuum is used. Default 0.

        mode: ``'cpu'``, ``'gpu'``
            if set to 'cpu', computes the spectra purely on the CPU. if set to 'gpu',
            offloads the calculations of lineshape and broadening steps to the GPU
            making use of parallel computations to speed up the process. Default 'cpu'.
            Note that mode='gpu' requires CUDA compatible hardware to execute. For more information on how to setup your system to run GPU-accelerated methods using CUDA and Cython, check `GPU Spectrum Calculation on RADIS <https://radis.readthedocs.io/en/latest/lbl/gpu.html>`
    ​
        Returns
        -------

        s: :class:`~radis.spectrum.spectrum.Spectrum`
            Output spectrum.
    ​
            Use the :py:meth:`~radis.spectrum.spectrum.Spectrum.get` method to retrieve a
            spectral quantity (``'radiance'``, ``'radiance_noslit'``, ``'absorbance'``, etc...)
    ​
            Or the :py:meth:`~radis.spectrum.spectrum.Spectrum.plot` method to plot it
            directly.
    ​
            See [1]_ to get an overview of all Spectrum methods
    ​
        References
        ----------
    ​
        .. [1] RADIS doc: `Spectrum how to? <https://radis.readthedocs.io/en/latest/spectrum/spectrum.html#label-spectrum>`__
    ​    .. [2] RADIS GPU support: 'GPU Calculations on RADIS <https://radis.readthedocs.io/en/latest/lbl/gpu.html>'
    ​
        Examples
        --------
    ​
        Calculate a CO spectrum from the HITRAN database::
    ​
            s = calc_spectrum(1900, 2300,         # cm-1
                              molecule='CO',
                              isotope='1,2,3',
                              pressure=1.01325,   # bar
                              Tgas=1000,
                              mole_fraction=0.1,
                              )
            s.apply_slit(0.5, 'nm')
            s.plot('radiance')

        This example uses the :py:meth:`~radis.spectrum.spectrum.Spectrum.apply_slit`
        and :py:meth:`~radis.spectrum.spectrum.Spectrum.plot` methods. See also
        :py:meth:`~radis.spectrum.spectrum.Spectrum.line_survey`::

            s.line_survey(overlay='radiance')

        Calculate a CO2 spectrum from the CDSD-4000 database:

            s = calc_spectrum(2200, 2400,   # cm-1
                              molecule='CO2',
                              isotope='1',
                              databank='/path/to/cdsd/databank/in/npy/format/',
                              pressure=0.1,  # bar
                              Tgas=1000,
                              mole_fraction=0.1,
                              mode='gpu'
                              )

            s.plot('absorbance')

        This example uses the :py:meth:`~radis.lbl.factor.eq_spectrum_gpu` method to calculate
        the spectrum on the GPU. The databank points to the CDSD-4000 databank that has been
        pre-processed and stored in `numpy.npy` format.
    ​
        Refer to the online :ref:`Examples <label_examples>` for more cases, and to
        the :ref:`Spectrum page <label_spectrum>` for details on post-processing methods.

        For more details on how to use the GPU method and process the database, refer to the examples
        linked above and the documentation on :ref:`GPU support for RADIS <label_radis_gpu>`.
    ​
        See Also
        --------

        :class:`~radis.lbl.factory.SpectrumFactory`,
        the :ref:`Spectrum page <label_spectrum>`
    """

    from radis.los.slabs import MergeSlabs

    if molecule is not None and type(molecule) != list:
        molecule = [molecule]  # fall back to the other case: multiple molecules

    # Stage 1. Find all molecules, whatever the user input configuration

    # ... Input arguments that CAN be dictionaries of molecules.
    DICT_INPUT_ARGUMENTS = {
        "isotope": isotope,
        "mole_fraction": mole_fraction,
        "databank": databank,
    }
    # Same, but when the values of the arguments themselves are already a dict.
    # (dealt with separately because we cannot use them to guess what are the input molecules)
    DICT_INPUT_DICT_ARGUMENTS = {"overpopulation": overpopulation}

    def _check_molecules_are_consistent(
        molecule_reference_set, reference_name, new_argument, new_argument_name
    ):
        """Will test that molecules set are the same in molecule_reference_set and new_argument, if new_argument is a dict.
        molecule_reference_set is a set of molecules (yeah!).
        reference_name is the name of the argument from which we guessed the list of molecules (used to have a clear error message).
        new_argument is the new argument to check
        new_argument_name is its name

        Returns the set of molecules as found in new_argument, if applicable, else the molecule_reference_set (this allows us to parse all arguments too)

        Note that names are just here to provide clear error messages to the user if there is a contradiction.
        """

        if isinstance(new_argument, dict):
            if molecule_reference_set is None:  # input molecules are still unknown
                return set(new_argument.keys()), new_argument_name
            elif set(new_argument.keys()) != set(molecule_reference_set):
                raise ValueError(
                    "Keys of molecules in the {0} dictionary must be the same as given in `{1}=`, i.e: {2}. Instead, we got {3}".format(
                        new_argument_name,
                        reference_name,
                        molecule_reference_set,
                        set(new_argument.keys()),
                    )
                )
            else:
                return (
                    set(new_argument.keys()),
                    new_argument_name,
                )  # so now we changed the reference
        else:
            return molecule_reference_set, reference_name

    # Parse all inputs:
    molecule_reference_set = molecule
    reference_name = "molecule"
    for argument_name, argument in DICT_INPUT_ARGUMENTS.items():
        molecule_reference_set, reference_name = _check_molecules_are_consistent(
            molecule_reference_set, reference_name, argument, argument_name
        )

    # ... Now we are sure there are no contradctions. Just ensure we have molecules:
    if molecule_reference_set is None:
        raise ValueError(
            "Please enter the molecule(s) to calculate in the `molecule=` argument or as a dictionary in the following: {0}".format(
                list(DICT_INPUT_ARGUMENTS.keys())
            )
        )

    # Stage 2. Now we have the list of molecules. Let's get the input arguments for each of them.

    # ... Initialize and fill the master-dictionary
    molecule_dict = {}

    for molecule in molecule_reference_set:
        molecule_dict[molecule] = {}

        for argument_name, argument_dict in DICT_INPUT_ARGUMENTS.items():
            if isinstance(argument_dict, dict):
                # Choose the correspond value
                molecule_dict[molecule][argument_name] = argument_dict[molecule]
                # Will raise a KeyError if not defined. That's fine!
                # TODO: maybe need to catch KeyError and raise a better error message?
            else:  # argument_name is not a dictionary.
                # Let's distribute the same value to every molecule:
                molecule_dict[molecule][argument_name] = argument_dict
                # If wrong argument, it will be caught in _calc_spectrum() later.

    # ... Special case of dictionary arguments. Find out if they were given as default, or per dictionary of molecules:
    is_same_for_all_molecules = dict.fromkeys(DICT_INPUT_DICT_ARGUMENTS)
    for argument_name, argument_dict in DICT_INPUT_DICT_ARGUMENTS.items():
        if not isinstance(argument_dict, dict):
            is_same_for_all_molecules[argument_name] = True
        else:
            # Argument is a dictionary. Guess if keys are molecules, or levels.
            # Ex: overpopulation dict could be {'CO2':{'(0,0,0,1)':10}} or directly {{'(0,0,0,1)':10}}
            argument_keys = set(argument_dict.keys())
            if all_in(argument_keys, molecule_reference_set):
                is_same_for_all_molecules[argument_name] = False
            else:
                is_same_for_all_molecules[argument_name] = True
    # ... now fill them in:
    for argument_name, argument_dict in DICT_INPUT_DICT_ARGUMENTS.items():
        if is_same_for_all_molecules[argument_name]:
            for mol in molecule_reference_set:
                # copy the value for everyone
                molecule_dict[mol][argument_name] = deepcopy(
                    argument_dict
                )  # in case it gets edited.
        else:  # argument_dict keys are the molecules:
            for mol in molecule_reference_set:
                molecule_dict[mol][argument_name] = argument_dict[mol]

    # Stage 3: Now let's calculate all the spectra
    s_list = []
    for molecule, dict_arguments in molecule_dict.items():
        kwargs_molecule = deepcopy(
            kwargs
        )  # these are the default supplementary arguments. Deepcopy ensures that they remain the same for all molecules, even if modified in _calc_spectrum

        # We add all of the DICT_INPUT_ARGUMENTS values:
        kwargs_molecule.update(**dict_arguments)

        s_list.append(
            _calc_spectrum(
                wavenum_min=wavenum_min,
                wavenum_max=wavenum_max,
                wavelength_min=wavelength_min,
                wavelength_max=wavelength_max,
                Tgas=Tgas,
                Tvib=Tvib,
                Trot=Trot,
                pressure=pressure,
                # overpopulation=overpopulation,  # now in dict_arguments
                molecule=molecule,
                # isotope=isotope,                # now in dict_arguments
                # mole_fraction=mole_fraction,    # now in dict_arguments
                path_length=path_length,
                # databank=databank,              # now in dict_arguments
                medium=medium,
                wstep=wstep,
                broadening_max_width=broadening_max_width,
                optimization=optimization,
                name=name,
                use_cached=use_cached,
                verbose=verbose,
                mode=mode,
                **kwargs_molecule
            )
        )

    # Stage 4: merge all molecules and return
    return MergeSlabs(*s_list)


def _calc_spectrum(
    wavenum_min,
    wavenum_max,
    wavelength_min,
    wavelength_max,
    Tgas,
    Tvib,
    Trot,
    pressure,
    overpopulation,
    molecule,
    isotope,
    mole_fraction,
    path_length,
    databank,
    medium,
    wstep,
    broadening_max_width,
    optimization,
    name,
    use_cached,
    verbose,
    mode,
    **kwargs
):
    """See :py:func:`~radis.lbl.calc.calc_spectrum`"""

    # Check inputs

    # ... wavelengths / wavenumbers
    if (wavelength_min is not None or wavelength_max is not None) and (
        wavenum_min is not None or wavenum_max is not None
    ):
        raise ValueError("Wavenumber and Wavelength both given... it's time to choose!")

    if wavenum_min is None and wavenum_max is None:
        assert wavelength_max is not None
        assert wavelength_min is not None
        wavenum_min = nm2cm(wavelength_max)
        wavenum_max = nm2cm(wavelength_min)
    else:
        assert wavenum_min is not None
        assert wavenum_max is not None

    # ... temperatures

    if Tgas is None and Trot is None:
        raise ValueError(
            "Choose either Tgas (equilibrium) or Tvib / Trot (non equilibrium)"
        )

    if Tvib is None and Trot is not None or Tvib is not None and Trot is None:
        raise ValueError("Choose both Tvib and Trot")

    # ... others
    if databank is None:
        raise ValueError("Give a databank name")

    if not "save_memory" in kwargs:
        # no need to save intermediary results as
        # factory is used once only
        kwargs["save_memory"] = True

    if "chunksize" in kwargs:
        raise DeprecationWarning("use optimization= instead of chunksize=")

    def _is_at_equilibrium():
        try:
            assert Tvib is None or Tvib == Tgas
            assert Trot is None or Trot == Tgas
            assert overpopulation is None
            if "self_absorption" in kwargs:
                assert kwargs["self_absorption"]  # == True
            return True
        except AssertionError:
            return False

    _equilibrium = _is_at_equilibrium()

    # which columns to keep when loading line database
    if kwargs["save_memory"] >= 2 and _equilibrium:
        drop_columns = "all"
    else:
        drop_columns = "auto"

    # Run calculations
    sf = SpectrumFactory(
        wavenum_min,
        wavenum_max,
        medium=medium,
        molecule=molecule,
        isotope=isotope,
        pressure=pressure,
        wstep=wstep,
        broadening_max_width=broadening_max_width,
        db_use_cached=use_cached,
        verbose=verbose,
        optimization=optimization,
        **kwargs
    )
    if databank == "fetch":  # mode to get databank without relying on  Line databases
        if _equilibrium:
            # Get line database from HITRAN
            # and partition functions from HAPI
            sf.fetch_databank(
                source="astroquery",
                format="hitran",
                parfuncfmt="hapi",  # use HAPI partition functions for equilibrium
                levelsfmt=None,  # no need to load energies
            )
        else:
            # Also get line database from HITRAN, and calculate partition functions
            # with energy levels from built-in constants (not all molecules
            # are supported!)
            sf.fetch_databank(
                source="astroquery",
                format="hitran",
                parfuncfmt="hapi",  # use HAPI partition functions for equilibrium
                levelsfmt="radis",  # built-in spectroscopic constants
            )
    elif exists(databank):
        # Guess format
        if databank.endswith(".par"):
            if verbose:
                print("Infered {0} is a HITRAN-format file.".format(databank))
            # If non-equilibrium we'll also need to load the energy levels.
            if _equilibrium:
                # Get partition functions from HAPI
                sf.load_databank(
                    path=databank,
                    format="hitran",
                    parfuncfmt="hapi",  # use HAPI partition functions for equilibrium
                    levelsfmt=None,  # no need to load energies
                    drop_columns=drop_columns,
                )
            else:
                # calculate partition functions with energy levels from built-in
                # constants (not all molecules are supported!)
                sf.load_databank(
                    path=databank,
                    format="hitran",
                    parfuncfmt="hapi",  # use HAPI partition functions for equilibrium
                    levelsfmt="radis",  # built-in spectroscopic constants
                    drop_columns=drop_columns,
                )
        elif databank.endswith(".npy"):
            if verbose:
                print("Infered {0} is a NPY-format file".format(databank))

            if _equilibrium:
                sf.load_databank(
                    path=databank,
                    format="cdsd-hitemp",
                    parfuncfmt="hapi",
                    levelsfmt=None,
                    buffer="npy",
                )
            else:
                raise (
                    AttributeError(
                        "Non equilibirum spectra calculation not yet supported with npy databank"
                    )
                )
        else:
            raise ValueError(
                "Couldnt infer the format of the line database file: {0}. ".format(
                    databank
                )
                + "Create a user-defined database in your ~/.radis file "
                + "and define the format there. More information on "
                + "https://radis.readthedocs.io/en/latest/lbl/lbl.html#configuration-file"
            )

    else:  # manual mode: get from user-defined line databases defined in ~/.radis
        sf.load_databank(
            databank,
            load_energies=not _equilibrium,  # no need to load/calculate energies at eq.
            drop_columns=drop_columns,
        )

    #    # Get optimisation strategies
    #    if lineshape_optimization == 'auto':        # NotImplemented: finally we use DLM all the time as default.
    #        if len(sf.df0) > 1e5:
    #            lineshape_optimization = 'DLM'
    #        else:
    #            lineshape_optimization = None
    #        sf.params['chunksize'] = lineshape_optimization

    # Use the standard eq_spectrum / non_eq_spectrum functions
    if _equilibrium:
        if mode == "cpu":
            s = sf.eq_spectrum(
                Tgas=Tgas,
                mole_fraction=mole_fraction,
                path_length=path_length,
                name=name,
            )
        else:
            s = sf.eq_spectrum_gpu(
                Tgas=Tgas,
                mole_fraction=mole_fraction,
                pressure=pressure,
                path_length=path_length,
                name=name,
            )
    else:
        s = sf.non_eq_spectrum(
            Tvib=Tvib,
            Trot=Trot,
            Ttrans=Tgas,
            overpopulation=overpopulation,
            mole_fraction=mole_fraction,
            path_length=path_length,
            name=name,
        )

    return s


# --------------------------
if __name__ == "__main__":

    from radis.test.lbl.test_calc import _run_testcases

    print(_run_testcases(verbose=True))



# -*- coding: utf-8 -*-
"""
Define warnings for radiation code, and how to deal with them

----------


"""


import warnings

from radis.misc.printer import printr

# %% Spectrum warnings / errors
# -----------------------------


class SlitDispersionWarning(UserWarning):
    """Warning trigger if Slit dispersion is too large"""


# %% Spectrum Factory warnings / errors
# -------------------------------------


class OutOfBoundError(ValueError):
    pass


class OutOfBoundWarning(UserWarning):
    """ Out of bound (for partition functions)"""

    pass


# Errors


class EmptyDatabaseError(ValueError):
    pass


# Warnings


class GaussianBroadeningWarning(UserWarning):
    pass


class CollisionalBroadeningWarning(UserWarning):
    pass


class VoigtBroadeningWarning(UserWarning):
    pass


class MemoryUsageWarning(UserWarning):
    pass


class EmptyDatabaseWarning(UserWarning):
    """ Trigger a warning if Line database is empty in the range considered """

    pass


class OutOfRangeLinesWarning(UserWarning):
    """Trigger a warning if out of range neighbouring lines, that could have
    an effect on the spectrume due to their broadening, cannot be found in the
    database"""

    pass


class HighTemperatureWarning(UserWarning):
    """Warning triggered when the Line database seems inappropriate for the
    temperatures considered
    """

    pass


class NegativeEnergiesWarning(UserWarning):
    pass


class MissingSelfBroadeningWarning(UserWarning):
    """Self broadening is missing in Line Database. Usually, use Air broadening
    instead"""

    pass


class LinestrengthCutoffWarning(UserWarning):
    """Warning triggered when the cumulated linestrength after intensity cutoff
    has changed too much"""

    pass


class InputConditionsWarning(UserWarning):
    """ Warning triggered when Spectrum input conditions are suspicious """

    pass


class PerformanceWarning(UserWarning):
    """ Warning triggered when it seems computation parameters are not optimized"""

    pass


WarningClasses = {
    "default": UserWarning,
    "SlitDispersionWarning": SlitDispersionWarning,
    "GaussianBroadeningWarning": GaussianBroadeningWarning,
    "CollisionalBroadeningWarning": CollisionalBroadeningWarning,
    "VoigtBroadeningWarning": VoigtBroadeningWarning,
    "MemoryUsageWarning": MemoryUsageWarning,
    "EmptyDatabaseWarning": EmptyDatabaseWarning,
    "OutOfRangeLinesWarning": OutOfRangeLinesWarning,
    "HighTemperatureWarning": HighTemperatureWarning,
    "NegativeEnergiesWarning": NegativeEnergiesWarning,
    "MissingSelfBroadeningWarning": MissingSelfBroadeningWarning,
    "LinestrengthCutoffWarning": LinestrengthCutoffWarning,
    "InputConditionsWarning": InputConditionsWarning,
    "PerformanceWarning": PerformanceWarning,
    "OutOfBoundWarning": OutOfBoundWarning,
}
""" dict: warnings used in RADIS Spectrum calculations.

You can selectively activate them by setting the warnings attribute of
:class:`radis.lbl.factory.SpectrumFactory`

See Also
--------

:py:data:`~radis.misc.warning.default_warning_status`
"""

# Setup individual warnings. Value of keys can be:
# - 'warning' (default: just trigger a warning)
# - 'error' (raises an error on this warning)
# - 'ignore'  (do nothing)
# The key self.warnings['default'] will set the warning behavior for all
# other warnings
default_warning_status = {
    "default": "warn",  # default
    "SlitDispersionWarning": "warn",
    "GaussianBroadeningWarning": "once",  # once per Spectrum calculation
    "CollisionalBroadeningWarning": "once",  # once per Spectrum calculation
    "VoigtBroadeningWarning": "once",  # once per Spectrum calculation
    # see also self.misc.warning_broadening_threshold for the treshold value
    "LinestrengthCutoffWarning": "warn",
    "MemoryUsageWarning": "warn",
    "EmptyDatabaseWarning": "warn",
    "OutOfRangeLinesWarning": "warn",
    "HighTemperatureWarning": "warn",
    "NegativeEnergiesWarning": "warn",  # warning if negative energies in database
    # warning if self-broadening abs coefficnet missing (Air is used instead)
    "MissingSelfBroadeningWarning": "warn",
    "InputConditionsWarning": "warn",
    "PerformanceWarning": "warn",
    "OutOfBoundWarning": "warn",
}
""" dict: default status of warnings used in RADIS Spectrum calculations.

Value of keys can be:

- 'warning' (default: just trigger a warning)
- 'error' (raises an error on this warning)
- 'ignore'  (do nothing)

The key self.warnings['default'] will set the warning behavior for all
other warnings. All warnings can be disabled by setting the SpectrumFactory
:py:attr:`~radis.lbl.loader.DatabankLoader.warnings` attribute to ``False``.

See Also
--------

:py:data:`~radis.misc.warning.WarningClasses`,
:py:func:`~radis.misc.warning.reset_warnings`

"""


def reset_warnings(status):
    """Reactivate warnings that are set 'once' per session in the Factory
    (unless all warnings have been set to False)

    Parameters
    ----------

    status: dict
        dictionary of Warnings with associated status

    """

    if status == False:
        return

    for k, v in status.items():
        if v == "once":
            WarningType = WarningClasses[k]
            warnings.simplefilter("default", WarningType)


def warn(message, category="default", status={}):
    """Trigger a warning, an error or just ignore based on the value defined
    in the :py:attr:`~radis.lbl.loader.DatabankLoader.warnings` dictionary

    The warnings can thus be deactivated selectively by setting the SpectrumFactory
    :attr:`~radis.lbl.loader.DatabankLoader.warnings` attribute. All warnings
    can be disabled by setting it to ``False``.

    Parameters
    ----------

    message: str
        what to print

    category: str
        one of the keys of self.warnings.

    status: dict
        status for all warning categories. Can be one of ``'warn'``, ``'ignore'``,
        ``'print'``, ``'error'``

    """

    if status == False:
        return

    action = status[category]

    WarningType = WarningClasses[category]

    if action in "warn":
        warnings.warn(WarningType(message))
    elif action == "once":
        warnings.warn(WarningType(message))
        # keep 'once' but ignore WarningType with simplefilters
        warnings.simplefilter("ignore", WarningType)
    elif action == "ignore":
        pass
    elif action == "print":  # just print the message, in red
        printr(message)
    elif action == "error":
        raise WarningType(message)
    else:
        raise ValueError("Unexpected action for warning: {0}".format(action))


# Tests (on module load)


# ... test warnings are well defined
for k in default_warning_status.keys():
    assert k in WarningClasses

# ... and reciprocally, but they all have a default value
for k in WarningClasses.keys():
    assert k in default_warning_status



# -*- coding: utf-8 -*-
""" Test parsers

Notes
-----

Runs tests for radis/io so that they can be accessed by pytest (and hopefully
the CI test suite)

Examples
--------

Run all tests::

	pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that have a 'fast' label)::

	pytest -m fast

-------------------------------------------------------------------------------

"""

import os
from os.path import getmtime
from warnings import warn

import numpy as np
import pytest

from radis.io.cdsd import cdsd2df
from radis.io.hitran import hit2df
from radis.test.utils import getTestFile, setup_test_line_databases


@pytest.mark.fast
def test_hitran_names_match(verbose=True, warnings=True, *args, **kwargs):
    """Compare that HITRAN species defined in :mod:`radis.io.hitran` match
    the nomenclature dictionary : :py:data:`radis.io.hitran.trans`.

    This should be ensured by developers when adding new species.
    """
    from radis.db.classes import (
        HITRAN_CLASS1,
        HITRAN_CLASS2,
        HITRAN_CLASS3,
        HITRAN_CLASS4,
        HITRAN_CLASS5,
        HITRAN_CLASS6,
        HITRAN_CLASS7,
        HITRAN_CLASS8,
        HITRAN_CLASS9,
        HITRAN_CLASS10,
        HITRAN_MOLECULES,
    )
    from radis.misc.basics import compare_lists

    all_hitran = (
        HITRAN_CLASS1
        + HITRAN_CLASS2
        + HITRAN_CLASS3
        + HITRAN_CLASS4
        + HITRAN_CLASS5
        + HITRAN_CLASS6
        + HITRAN_CLASS7
        + HITRAN_CLASS8
        + HITRAN_CLASS9
        + HITRAN_CLASS10
    )
    all_hitran = list(set(all_hitran))

    # All species in HITRAN groups should be valid HITRAN_MOLECULES names
    for m in all_hitran:
        if not m in HITRAN_MOLECULES:
            raise ValueError(
                "{0} is defined in HITRAN groups but has no HITRAN id".format(m)
            )

    # Species in 'HITRAN_MOLECULES' should be classified in groups, else nonequilibrium
    # calculations are not possible.
    if warnings and all_hitran != HITRAN_MOLECULES:
        warn(
            "Difference between HITRAN groups (left) and HITRAN id "
            + "dictionary (right). Some HITRAN species are not classified in "
            + "groups. Nonequilibrium calculations wont be possible for these!:\n"
            + "{0}".format(
                compare_lists(
                    all_hitran, HITRAN_MOLECULES, verbose=False, return_string=True
                )[1]
            )
        )
    return


@pytest.mark.fast
def test_hitran_co(verbose=True, warnings=True, **kwargs):
    """ Analyse some default files to make sure everything still works"""

    # 1. Load
    df = hit2df(getTestFile("hitran_CO_fragment.par"), cache="regen")
    if verbose:
        print("Read hitran_CO_fragment.par")
        print("---------------------------")
        print(df.head())

    # 2. Test
    assert list(df.loc[0, ["vu", "vl"]]) == [4, 4]
    assert df.dtypes["vu"] == np.int64
    assert df.dtypes["vl"] == np.int64

    return True


def test_hitran_co2(verbose=True, warnings=True, **kwargs):

    # 1. Load
    df = hit2df(getTestFile("hitran_CO2_fragment.par"), cache="regen")
    if verbose:
        print("Read hitran_CO2_fragment.par")
        print("----------------------------")
        print(df.head())

    # 2. Test
    assert list(
        df.loc[0, ["v1u", "v2u", "l2u", "v3u", "v1l", "v2l", "l2l", "v3l"]]
    ) == [4, 0, 0, 0, 0, 0, 0, 1]
    assert df.dtypes["v1l"] == np.int64
    assert df.dtypes["v3u"] == np.int64

    return True


def test_hitran_h2o(verbose=True, warnings=True, **kwargs):

    # 1. Load
    df = hit2df(getTestFile("hitran_2016_H2O_2iso_2000_2100cm.par"), cache="regen")
    if verbose:
        print("Read hitran_2016_H2O_2iso_2000_2100cm.par")
        print("-----------------------------------------")
        print(df.head())

    # 2. Test
    assert list(df.loc[0, ["v1u", "v2u", "v3u", "v1l", "v2l", "v3l"]]) == [
        0,
        2,
        0,
        0,
        1,
        0,
    ]

    assert df.loc[26, "ju"] == 5  # in .par : line 27, column 99-100
    assert df.loc[27, "ju"] == 18  # in .par : line 28, column 99-100

    assert df.dtypes["v1l"] == np.int64
    assert df.dtypes["v3u"] == np.int64

    assert df.dtypes["ju"] == np.int64
    assert df.dtypes["Kau"] == np.int64
    assert df.dtypes["Kcu"] == np.int64
    assert df.dtypes["jl"] == np.int64
    assert df.dtypes["Kal"] == np.int64
    assert df.dtypes["Kcl"] == np.int64

    return True


def test_hitemp(verbose=True, warnings=True, **kwargs):
    """ Analyse some default files to make sure everything still works"""

    # 1. Load
    df = cdsd2df(
        getTestFile("cdsd_hitemp_09_header.txt"), cache="regen", drop_non_numeric=True
    )
    if verbose:
        print(df.head())

    # 2. Tests
    assert df.wav[3] == 2250.00096
    # make sure P Q R is correctly replaced by drop_non_numeric:
    assert "branch" in df
    assert df["branch"].iloc[0] == 0  # Q
    assert df["branch"].iloc[1] == 1  # R

    return True


def run_example():
    from radis import SpectrumFactory

    setup_test_line_databases(
        verbose=True
    )  # add HITEMP-CO2-HAMIL-TEST in ~/.radis if not there
    sf = SpectrumFactory(
        wavelength_min=4165,
        wavelength_max=4200,
        path_length=0.1,
        pressure=20,
        molecule="CO2",
        isotope="1,2",
        cutoff=1e-25,  # cm/molecule
        broadening_max_width=10,  # cm-1
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.load_databank("HITRAN-CO2-TEST")  # this database must be defined in ~/.radis


def test_cache_regeneration(verbose=True, warnings=True, **kwargs):

    run_example()
    # get time when line database file was last modified
    file_last_modification = getmtime(
        getTestFile(r"hitran_co2_626_bandhead_4165_4200nm.par")
    )
    # get time when cache file was last modified
    cache_last_modification = getmtime(
        getTestFile(r"hitran_co2_626_bandhead_4165_4200nm.h5")
    )
    # change the time when line database was last modified
    stinfo = os.stat(getTestFile(r"hitran_co2_626_bandhead_4165_4200nm.par"))
    access_time = stinfo.st_atime
    os.utime(
        getTestFile(r"hitran_co2_626_bandhead_4165_4200nm.par"),
        (file_last_modification + 1, access_time + 1),
    )

    file_last_modification_again = getmtime(
        getTestFile(r"hitran_co2_626_bandhead_4165_4200nm.par")
    )

    assert file_last_modification_again > file_last_modification

    run_example()

    cache_last_modification_again = getmtime(
        getTestFile(r"hitran_co2_626_bandhead_4165_4200nm.h5")
    )

    assert cache_last_modification_again > cache_last_modification


def _run_testcases(verbose=True, *args, **kwargs):

    test_hitran_names_match(verbose=verbose, *args, **kwargs)
    test_hitran_co(verbose=verbose, *args, **kwargs)
    test_hitran_co2(verbose=verbose, *args, **kwargs)
    test_hitran_h2o(verbose=verbose, *args, **kwargs)
    test_hitemp(verbose=verbose, *args, **kwargs)

    test_cache_regeneration(verbose=verbose, *args, **kwargs)
    return True


if __name__ == "__main__":
    print("Testing io.py: ", _run_testcases(verbose=True))



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 20 09:38:23 2017

@author: erwan

Use
------

Run all tests:

>>> pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that a 'fast' label)

>>> pytest -m fast

"""

from time import time
from warnings import warn

from radis.lbl.parallel import ParallelFactory, SpectrumFactory
from radis.misc.printer import printm
from radis.test.utils import setup_test_line_databases

# %% Main


def test_parallel(plot=False, verbose=True, warnings=True, *args, **kwargs):
    """Core of the parallel test routine

    Test than parallel is at least faster than sequential mode. Note that
    this is not necessary: parallelization becomes much faster during the
    convolution steps, hence for large number of lines
    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        import matplotlib.pyplot as plt

        plt.ion()

    setup_test_line_databases()

    # Performance test
    Tvib = 1000  # , 1200, 1400]
    Trot = [900, 1200, 1250, 1300, 1400, 1600]
    path_length = 0.1
    mole_fraction = 0.5

    cutoff = 0
    broadening_max_width = 50
    wstep = 0.01

    # .. parallel
    pf = ParallelFactory(
        wavelength_min=4165,
        wavelength_max=4200,
        cutoff=cutoff,
        db_use_cached=True,
        lvl_use_cached=True,
        isotope="1",
        broadening_max_width=broadening_max_width,
        wstep=wstep,
        verbose=verbose,
    )
    pf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    pf.warnings["NegativeEnergiesWarning"] = "ignore"
    pf.warnings["EmptyDatabaseWarning"] = "ignore"
    pf.load_databank("HITRAN-CO2-TEST")

    t0 = time()
    s_parallel = pf.non_eq_spectrum(
        Tvib, Trot, path_length=path_length, mole_fraction=mole_fraction
    )
    t0 = time() - t0

    # ... non parallel
    sf = SpectrumFactory(
        wavelength_min=4165,
        wavelength_max=4200,
        cutoff=cutoff,
        db_use_cached=True,
        lvl_use_cached=True,
        isotope="1,2",
        broadening_max_width=broadening_max_width,
        wstep=wstep,
        verbose=verbose,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.warnings["NegativeEnergiesWarning"] = "ignore"
    sf.warnings["EmptyDatabaseWarning"] = "ignore"
    sf.load_databank("HITRAN-CO2-TEST")
    t1 = time()
    s_sequential = []
    for Troti in Trot:
        s_sequential.append(
            sf.non_eq_spectrum(
                Tvib, Troti, path_length=path_length, mole_fraction=mole_fraction
            )
        )
    t1 = time() - t1

    if verbose:
        printm(("\nParallelFactory performance test ({0} cases)".format(len(Trot))))
        printm("-------------------")
        printm(("Parallel version: {0:.1f}s".format(t0)))
        printm(("Normal version: {0:.1f}s".format(t1)))
        printm("-------------------")

    # Assert that spectra are the same
    for i in range(len(Trot)):
        assert s_parallel[i].compare_with(
            s_sequential[i], spectra_only=True, plot=False
        )
    if verbose:
        printm("Ensured that spectra are the same: OK")

    #        # parallel version should be at least faster!
    #        # note: ...  may not be the case if the calculated ranges are too small
    #        # ......... as is the case in the test database. Better use // for spectra
    #        # ......... with calculation times > 5s
    #        assert t0 > t1
    if t0 > t1:
        warn(
            "Parallel version ({0:.1f}s) is slower than the Serial version ({1:.1f}s) !".format(
                t0, t1
            )
        )

    return True  # no plot implemented


def DISCARDED_test_parallel_internal(
    plot=False, verbose=True, warnings=True, *args, **kwargs
):
    """Core of the parallel test routine

    Test than parallel is at least faster than sequential mode. Note that
    this is not necessary: parallelization becomes much faster during the
    convolution steps, hence for large number of lines
    """
    # TODO: there is a line missing in the parallel case at 2283nm.
    # I expect the merging of different wave ranges to fail.
    # Test is not run for the moment

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        import matplotlib.pyplot as plt

        plt.ion()

    setup_test_line_databases()

    # Performance test
    Tvib = 1000  # , 1200, 1400]
    Trot = 900
    path_length = 0.1
    mole_fraction = 0.5

    cutoff = 0
    broadening_max_width = 1
    wstep = 0.0001

    # .. parallel
    sfp = SpectrumFactory(
        wavelength_min=4377.134,
        wavelength_max=4377.9,
        pressure=0.020,  # bar
        cutoff=cutoff,
        db_use_cached=True,
        lvl_use_cached=True,
        parallel=True,
        isotope="1",
        broadening_max_width=broadening_max_width,
        wstep=wstep,
        verbose=verbose,
    )
    sfp.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sfp.warnings["NegativeEnergiesWarning"] = "ignore"
    sfp.warnings["EmptyDatabaseWarning"] = "ignore"
    sfp.load_databank("HITEMP-CO2-TEST")

    s_parallel = sfp.non_eq_spectrum(
        Tvib, Trot, path_length=path_length, mole_fraction=mole_fraction
    )
    s_parallel.name = "Parallel ({0:.1f}s)".format(
        s_parallel.conditions["calculation_time"]
    )

    # ... non parallel
    sf = SpectrumFactory(
        wavelength_min=4377.134,
        wavelength_max=4377.9,
        pressure=0.020,  # bar
        cutoff=cutoff,
        db_use_cached=True,
        lvl_use_cached=True,
        isotope="1,2",
        broadening_max_width=broadening_max_width,
        wstep=wstep,
        verbose=verbose,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.warnings["NegativeEnergiesWarning"] = "ignore"
    sf.warnings["EmptyDatabaseWarning"] = "ignore"
    sf.load_databank("HITEMP-CO2-TEST")
    s_sequential = sf.non_eq_spectrum(
        Tvib, Trot, path_length=path_length, mole_fraction=mole_fraction
    )
    s_sequential.name = "Sequential ({0:.1f}s)".format(
        s_sequential.conditions["calculation_time"]
    )

    t0 = s_parallel.conditions["calculation_time"]
    t1 = s_sequential.conditions["calculation_time"]

    if verbose:
        printm("\nSpectrumFactory with internal parallel test")
        printm("-------------------")
        printm(("Parallel version: {0:.1f}s".format(t0)))
        printm(("Normal version: {0:.1f}s".format(t1)))
        printm("-------------------")

    # Assert that spectra are the same
    assert s_parallel.compare_with(s_sequential, spectra_only=True, plot=False)
    if verbose:
        printm("Ensured that spectra are the same: OK")

    if t0 > t1:
        warn(
            "Parallel version ({0:.1f}s) is slower than the Serial version ({1:.1f}s) !".format(
                t0, t1
            )
        )

    return True  # no plot implemented


def _run_testcases(plot=False, verbose=True, *args, **kwargs):

    test_parallel(plot=plot, verbose=verbose, *args, **kwargs)


#    test_parallel_internal(plot=plot, verbose=verbose, *args, **kwargs)

if __name__ == "__main__":

    printm(("Tested parallel.py:", _run_testcases(plot=True)))



# -*- coding: utf-8 -*-
"""
Tools to deal with HDF5 cache files
HDF5 cache files are used to cache Energy Database files, and Line Database
files, and yield a much faster access time.

Routine Listing
---------------


- :func:`~radis.misc.cache_file.check_cache_file`
- :func:`~radis.misc.cache_file.check_not_deprecated`
- :func:`~radis.misc.cache_file.save_to_hdf`

See Also
--------

:py:func:`~radis.io.hitran.hit2df`,
:py:func:`~radis.io.cdsd.cdsd2df`

-------------------------------------------------------------------------------


"""
# TODO: start using .feather format. Faster than HDF5 for pandas Dataframe,
# and can be multithreaded.
# see:
# https://gist.github.com/gansanay/4514ec731da1a40d8811a2b3c313f836
# and pd.read_feather(file, nthreads=3)


# Note: don't import unicode_literals because it breaks the df.to_hdf of
# save_to_hdf because of a stupid unicode/str error in Python 2.7
import os
from os.path import exists, splitext
from warnings import warn

import h5py
import pandas as pd
from packaging.version import parse

import radis
from radis import OLDEST_COMPATIBLE_VERSION
from radis.misc.basics import compare_dict, is_float
from radis.misc.printer import printm, printr


class DeprecatedFileError(DeprecationWarning):
    pass


"""str: forces to regenerate cache files that were created in a previous version"""

# Just make sure LAST_BACKWARD_COMPATIBLE_VERSION is valid
assert radis.__version__ >= OLDEST_COMPATIBLE_VERSION

# Utils


def load_h5_cache_file(
    cachefile,
    use_cached,
    metadata,
    current_version,
    last_compatible_version=OLDEST_COMPATIBLE_VERSION,
    verbose=True,
):
    """Function to load a h5 cache file

    Parameters
    ----------

    cachefile: str
        cache file path

    use_cached: str
        use cache file if value is not ``False``:

        - if ``True``, use (and generate if doesnt exist) cache file.
        - if ``'regen'``, delete cache file (if exists) so it is regenerated
        - if ``'force'``, use cache file and raises an error if it doesnt exist

        if using the cache file, check if the file is deprecated. If it
        is deprecated, regenerate the file unless ``'force'`` was used
        (in that case, raise an error)

    metadata: dict
        values are compared to cache file attributes. If they dont match,
        the file is considered deprecated. See ``use_cached`` to know
        how to handle deprecated files

    current_version: str
        version is compared to cache file version (part of attributes).
        If current version is superior, a simple warning is triggered.

    last_compatible_version: str
        if file version is inferior to this, file is considered deprecated.
        See ``use_cached`` to know how to handle deprecated files.
        Default :data:`~radis.OLDEST_COMPATIBLE_VERSION`.

    Returns
    -------

    df: pandas DataFrame, or None
        None if no cache file was found, or if it was deleted


    See Also
    --------

    :data:`~radis.OLDEST_COMPATIBLE_VERSION`

    """

    # 1. know if we have to load the file
    if not use_cached:
        return None
    elif use_cached == "regen" and exists(cachefile):
        os.remove(cachefile)
        if verbose:
            printm("Deleted h5 cache file : {0}".format(cachefile))
        return None

    # 2. check the file is here
    if not exists(cachefile):
        if use_cached == "force":
            raise ValueError("Cache file {0} doesnt exist".format(cachefile))
        else:
            return None  # File doesn't exist. It's okay.

    # 3. read file attributes to know if it's deprecated
    try:
        check_not_deprecated(
            cachefile,
            metadata,
            current_version=current_version,
            last_compatible_version=last_compatible_version,
        )
    # ... if deprecated, raise an error only if 'force'
    except DeprecatedFileError as err:
        if use_cached == "force":
            raise err
        else:
            if verbose:
                printr(
                    "File {0} deprecated:\n{1}\nDeleting it!".format(
                        cachefile, str(err)
                    )
                )
            os.remove(cachefile)
            return None
    # 4. File is not not deprecated: read the content.
    else:
        df = None
        if verbose >= 2:
            printm("Reading cache file ({0})".format(cachefile))
        try:
            df = pd.read_hdf(cachefile, "df")
        except KeyError as err:  # An error happened during file reading.
            # Fail safe by deleting cache file (unless we explicitely wanted it
            # with 'force')
            if use_cached == "force":
                raise
            else:
                if verbose:
                    printr(
                        "An error happened during cache file reading "
                        + "{0}:\n{1}\n".format(cachefile, str(err))
                        + "Deleting cache file to regenerate it"
                    )
                os.remove(cachefile)
                df = None

    return df


def get_cache_file(fcache, verbose=True):
    """Load HDF5 cache file

    Parameters
    ----------

    fcache: str
        file name

    Other Parameters
    ----------------

    verbose: bool
        If >=2, also warns if non numeric values are present (it would make
        calculations slower)

    Notes
    -----

    we could start using FEATHER format instead. See notes in cache_files.py
    """

    # Load file

    df = pd.read_hdf(fcache, "df")

    # Check file

    # ... 'object' columns slow everything down (not fixed format strings!)
    if verbose >= 2:
        _warn_if_object_columns(df, fcache)

    return df


def check_cache_file(fcache, use_cached=True, metadata={}, verbose=True):
    """Quick function that check status of cache file generated by RADIS:

    Parameters
    ----------

    fcache: str
        cache file name

    use_cached: ``True``, ``False``, ``'force'``, ``'regen'``
        see notes below. Default ``True``.

    metadata: dict
        attributes to check

    verbose: boolean
        print stuff

    Notes
    -----

    The function first checks the existence of ``fcache``. What is does depends
    on the value of ``use_cached``:

    - if ``True``, check it exists.
    - if ``'regen'``, delete cache file to regenerate it later.
    - if ``'force'``, raise an error if file doesnt exist.

    Then look if it is deprecated (we just look at the attributes, the file
    is never fully read). Deprecation is done by :py:func:`~radis.misc.cache_files.check_not_deprecated`
    comparing the ``metadata=`` content.

    - if deprecated, deletes it to regenerate later unless 'force' was used

    See Also
    --------

    :py:func:`~radis.misc.cache_files.check_not_deprecated`
    """

    # Test existence of file:
    if not use_cached:
        return  # we dont want a cache file, no need to test it
    elif use_cached == "regen":
        if exists(fcache):
            os.remove(fcache)
            if verbose:
                print(("Deleted h5 cache file : {0}".format(fcache)))
    elif use_cached == "force":
        if not exists(fcache):
            raise ValueError("Cache file {0} doesnt exist".format(fcache))
    else:  # use_cached == True
        pass  # just use the file as is

    # If file is still here, test if it is deprecated:
    # (we just read the attributes, the file is never fully read)
    if exists(fcache):
        if verbose:
            print(("Using cache file: {0}".format(fcache)))
        try:
            check_not_deprecated(
                fcache,
                metadata=metadata,
                current_version=radis.__version__,
                last_compatible_version=OLDEST_COMPATIBLE_VERSION,
            )
        except DeprecatedFileError as err:
            if use_cached == "force":
                raise
            else:  # delete file to regenerate it in the end of the script
                if verbose:
                    printr(
                        "File {0} deprecated:\n{1}\nDeleting it!".format(
                            fcache, str(err)
                        )
                    )
                os.remove(fcache)

    return


def check_not_deprecated(
    file,
    metadata,
    current_version=None,
    last_compatible_version=OLDEST_COMPATIBLE_VERSION,
):
    """Make sure cache file is not deprecated: checks that ``metadata`` is the same,
    and that the version under which the file was generated is valid.

    Parameters
    ----------

    file: str
        a `` .h5`` cache file for Energy Levels

    metadata: dict
        list of variables used to create the file. If the values dont match,
        an error is raised.

    current_version: str, or ``None``
        current version number. If the file was generated in a previous version
        a warning is raised. If ``None``, current version is read from
        :data:`radis.__version__`.

    last_backward_compatible_version: str
        If the file was generated in a non-compatible version, an error is raised.
        Default :py:data:`~radis.OLDEST_COMPATIBLE_VERSION`

    """

    # Get attributes (metadata+version)
    hf = h5py.File(file, "r")
    try:
        attrs = dict(hf.attrs)
    except OSError:
        attrs = {}
    finally:
        hf.close()

    # Raise an error if version is not found
    try:
        file_version = attrs.pop("version")
    except KeyError:
        raise DeprecatedFileError(
            "File {0} has been generated in a deprecated ".format(file)
            + "version. Delete it to regenerate it on next run"
        )

    # Get current version
    if current_version is None:
        current_version = radis.__version__

    # If file version is anterior to a major change
    # ... Update here versions afterwhich Deprecated cache file is not safe
    # ... (example: a key name was changed)
    if parse(file_version) < parse(last_compatible_version):
        raise DeprecatedFileError(
            "File {0} has been generated in a deprecated ".format(file)
            + "version ({0}). Oldest compatible version is {1}. ".format(
                file_version, last_compatible_version
            )
            + "Delete the file to regenerate it on next run"
        )

    # If file version is outdated: Warning, but no error
    if parse(current_version) > parse(file_version):
        warn(
            DeprecationWarning(
                "File {0} has been generated in ".format(file)
                + "a deprecated version ({0}) compared to current ({1})".format(
                    file_version, current_version
                )
                + ". Delete it to regenerate it on next run"
            )
        )
        out = False
    elif parse(current_version) == parse(file_version):
        out = True
    else:
        raise ValueError(
            "Cache file ({0}) generated with a future version ({1} > {2})? ".format(
                file, file_version, current_version
            )
            + "Do you own a DeLorean? Delete the file manually if you understand what happened"
        )

    # Compare metadata
    metadata = _h5_compatible(metadata)
    out, compare_string = compare_dict(
        metadata, attrs, verbose=False, return_string=True
    )
    if out != 1:
        raise DeprecatedFileError(
            "Metadata in file {0} dont match ".format(file)
            + "expected values. See comparison below:"
            + "\n\tExpected\tFile\n{0}".format(compare_string)
        )

    return out


def _h5_compatible(a_dict):
    """ Make dictionary ``a_dict`` h5 compatible """
    out = {}
    for k, v in a_dict.items():
        if v is None:
            continue  # dont store None
        elif is_float(v):
            out[k] = v
        else:
            out[k] = str(v)  # convert to str
    return out


def _warn_if_object_columns(df, fname):
    """  'object' columns slow everything down (not fixed format strings!) """
    objects = [k for k, v in df.dtypes.items() if v == object]
    if len(objects) > 0:
        warn(
            "Dataframe in {0} contains `object` format columns: {1}. ".format(
                fname, objects
            )
            + "Operations will be slower. Try to convert them to numeric."
        )


def save_to_hdf(
    df, fname, metadata, version=None, key="df", overwrite=True, verbose=True
):
    """Save energy levels to HDF5 file. Add metadata and version

    Parameters
    ----------

    df: a pandas DataFrame
        data will be stored in the key ``'df'``

    fname: str
        ``.h5`` file where to store.

    metadata: dict
         dictionary of values that were used to generate the DataFrame. Metadata
         will be asked again on file load to ensure it hasnt changed. ``None``
         values will not be stored.

    version: str, or ``None``
        file version. If ``None``, the current :data:`radis.__version__` is used.
        On file loading, a warning will be raised if the current version is
        posterior, or an error if the file version is set to be uncompatible.
        See :py:data:`~radis.OLDEST_COMPATIBLE_VERSION`

    key: str
        dataset name. Default ``'df'``

    overwrite: boolean
        if ``True``, overwrites file. Else, raise an error if it exists.

    Other Parameters
    ----------------

    verbose: bool
        If >=2, also warns if non numeric values are present (it would make
        calculations slower)

    Notes
    -----

    ``None`` values are not stored

    """

    # Check file
    assert fname.endswith(".h5")
    assert "version" not in metadata
    # ... 'object' columns slow everything down (not fixed format strings!)
    if verbose >= 2:
        _warn_if_object_columns(df, fname)

    # Update metadata format
    metadata = _h5_compatible(metadata)

    # Overwrite file
    if exists(fname) and not overwrite:
        raise ValueError("File exist: {0}".format(fname))
    hf = h5py.File(fname, "w")

    try:

        # Start by adding version
        if version is None:
            version = radis.__version__
        hf.attrs["version"] = version

        # Add metadata
        for k, v in metadata.items():
            hf.attrs[k] = v

        if verbose >= 3:
            print("... saved {0} with metadata: {1}".format(fname, metadata))

    except:
        raise

    finally:
        hf.close()

    # now export dataframe
    df.to_hdf(fname, key, format="fixed", mode="a", complevel=1, complib="blosc")


#    df.to_hdf(fname, 'df', format='fixed', mode='a')  # would be 10-20% faster, but take 2x more space


def filter_metadata(arguments, discard_variables=["self", "verbose"]):
    """Filter arguments (created with  ``locals()`` at the beginning
    of the script) to extract metadata. Metadata is stored as attributes
    in the cached file:

    - remove variables in ``discard_variables``
    - remove variables that start with ``'_'``
    - remove varibles whose value is ``None``

    Parameters
    ----------

    arguments: dict
        list of local variables. For instance::

            arguments = locals()

    discard_variables: list of str
        variable names to discard

    Returns
    -------

    metadata: dict
        a (new) dictionary built from arguments by removing ``discard_variables``
        and variables starting with ``'_'``

    Examples
    --------

    How to get only function argument::

        def some_function(*args):
            metadata = locals()     # stores only function arguments because it's the first line

            ...

            metadata = filter_metadata(metadata)
            save_to_hdf(df, fname, metadata=metadata)

            ...

    """

    metadata = {k: v for (k, v) in arguments.items() if not k.startswith("_")}
    metadata = {k: v for (k, v) in metadata.items() if k not in discard_variables}
    metadata = {k: v for (k, v) in metadata.items() if v is not None}

    return metadata


def cache_file_name(fname):
    """ Returns the corresponding cache file name for fname """
    return splitext(fname)[0] + ".h5"



# -*- coding: utf-8 -*-
"""
Summary
-------

Functions to parse the ``~/.radis`` :ref:`Configuration file <label_lbl_config_file>`

Notes
-----

Create a ~/.radis file in your HOME that contains all machine-specific information
(e.g: path to databanks). See :data:`~radis.misc.config.DBFORMAT` for expected
format

Routine Listing
---------------

- :func:`~radis.misc.config.getConfig`
- :func:`~radis.misc.config.getDatabankEntries`
- :func:`~radis.misc.config.getDatabankList`
- :func:`~radis.misc.config.addDatabankEntries`
- :func:`~radis.misc.config.diffDatabankEntries`
- :func:`~radis.misc.config.printDatabankEntries`
- :func:`~radis.misc.config.printDatabankList`



-------------------------------------------------------------------------------


"""


import configparser
import json
from os.path import dirname, exists, expanduser, join

from radis.misc.basics import compare_dict, compare_lists, stdpath
from radis.misc.utils import DatabankNotFound, getProjectRoot

# %% Functions to parse radis/config.json


def get_config():
    """ Read the config.json file """
    jsonfile = join(getProjectRoot(), "config.json")
    with open(jsonfile) as f:
        try:
            config = json.load(f)
        except json.JSONDecodeError as err:
            raise json.JSONDecodeError(
                "Error reading '{0}' (line {2} col {3}): \n{1}".format(
                    jsonfile, err.msg, err.lineno, err.colno
                ),
                err.doc,
                err.pos,
            ) from err
    return config


# %% Functions to parse ~/.radis file

DBFORMAT = r"""
--------------------------

[MY-HITEMP-CO2]                  #  your databank name: use this in calc_spectrum()
                                 #  or SpectrumFactory.load_databank()
info = HITEMP 2010 databank      #  whatever you want
path =                           #  no "", multipath allowed
       D:\Databases\HITEMP-CO2\hitemp_07
       D:\Databases\HITEMP-CO2\hitemp_08
       D:\Databases\HITEMP-CO2\hitemp_09
format = hitran                  #  'hitran' (HITRAN/HITEMP), 'cdsd-hitemp', 'cdsd-4000'
                                 # databank text file format. More info in
                                 # SpectrumFactory.load_databank function.
parfuncfmt:                      #  'cdsd', 'hapi', etc.
                                 # format to read tabulated partition function
                                 # file. If `hapi`, then HAPI (HITRAN Python
                                 # interface) is used to retrieve them (valid if
                                 # your databank is HITRAN data). HAPI is embedded
                                 # into RADIS. Check the version.
# Optional
# ----------
parfunc:                         #  path to tabulated partition function to use.
                                 # If `parfuncfmt` is `hapi` then `parfunc`
                                 # should be the link to the hapi.py file. If
                                 # not given, then the hapi.py embedded in RADIS
                                 # is used (check version)
levels_iso1                      #  path to energy levels (needed for non-eq
                                 # calculations). Default None
levels_iso2                      # etc
levels_iso4                      # etc
levelsfmt:                       #  'cdsd', etc.
                                 # how to read the previous file. Default None.
levelsZPE:                       #  zero-point-energy (cm-1): offset for all level
                                 # energies. Default 0 (if not given)

--------------------------"""
"""str: Typical expected format of a ~/.radis entry::

    --------------------------

    [MY-HITEMP-CO2]                  #  your databank name: use this in calc_spectrum()
                                     #  or SpectrumFactory.load_databank()
    info = HITEMP 2010 databank      #  whatever you want
    path =                           #  no "", multipath allowed
           D:\Databases\HITEMP-CO2\hitemp_07
           D:\Databases\HITEMP-CO2\hitemp_08
           D:\Databases\HITEMP-CO2\hitemp_09
    format = hitran                  #  'hitran' (HITRAN/HITEMP), 'cdsd-hitemp', 'cdsd-4000'
                                     # databank text file format. List of all
                                     # formats in :py:data:`~radis.lbl.loader.KNOWN_DBFORMAT`
                                     # More info in
                                     # :py:meth:`~radis.lbl.loader.DatabankLoader.load_databank` function.
    parfuncfmt:                      #  'cdsd', 'hapi', etc.
                                     # format to read tabulated partition function
                                     # file. If `hapi`, then HAPI (HITRAN Python
                                     # interface) is used to retrieve them (valid if
                                     # your databank is HITRAN data). HAPI is embedded
                                     # into RADIS. Check the version.
                                     # List of all formats in :py:data:`~radis.lbl.loader.KNOWN_LVLFORMAT`
    # Optional
    # ----------
    parfunc:                         #  path to tabulated partition function to use.
                                     # If `parfuncfmt` is `hapi` then `parfunc`
                                     # should be the link to the hapi.py file. If
                                     # not given, then the hapi.py embedded in RADIS
                                     # is used (check version)
    levels_iso1                      #  path to energy levels (needed for non-eq
                                     # calculations). Default None
    levels_iso2                      # etc
    levels_iso4                      # etc
    levelsfmt:                       #  'cdsd', etc.
                                     # how to read the previous file. Default None.
    levelsZPE:                       #  zero-point-energy (cm-1): offset for all level
                                     # energies. Default 0 (if not given)

    --------------------------

For more information refer to the documentation: :ref:`Configuration file <label_lbl_config_file>` :

- Setup test databases with :py:func:`~radis.test.utils.setup_test_line_databases`
- `format` :  format: :py:data:`~radis.lbl.loader.KNOWN_DBFORMAT`


See Also
--------

:ref:`Configuration file <label_lbl_config_file>`,
:py:func:`~radis.misc.config.getConfig`,
:py:meth:`~radis.lbl.loader.DatabankLoader.load_databank`
"""

CONFIG_PATH = join(expanduser("~"), ".radis")


def getConfig():
    """Read config file and returns it

    Config file name is harcoded: `~/.radis`
    """

    config = configparser.ConfigParser()
    configpath = CONFIG_PATH

    # Test ~/.radis exists
    if not exists(configpath):

        raise FileNotFoundError(
            "Create a `.radis` file in {0} to store links to ".format(
                dirname(configpath)
            )
            + "your local databanks. Format must be:\n {0}".format(DBFORMAT)
            + "\n(it can be empty too)"
        )
    config.read(configpath)

    return config


#


def getDatabankEntries(dbname):
    r"""Read ~/.radis config file and returns a dictionary of entries.


    Notes
    -----

    Databank format:

        [MY-HITEMP-CO2]                  #  your databank name: use this in calc_spectrum()
                                         #  or SpectrumFactory.load_databank()
        info = HITEMP 2010 databank      # whatever you want
        path =                           # no "", multipath allowed
               D:\Databases\HITEMP-CO2\hitemp_07
               D:\Databases\HITEMP-CO2\hitemp_08
               D:\Databases\HITEMP-CO2\hitemp_09
        format = hitemp                  # 'hitran' (HITRAN / HITEMP), 'cdsd-hitemp', 'cdsd-4000'
                                         # Databank text file format. More info in
                                         # SpectrumFactory.load_databank function.

        # Optional:

        parfunc                          # path or 'USE_HAPI'
                                         # path to tabulated partition functions. If
                                         # `USE_HAPI`, then HAPI (HITRAN Python
                                         interface) [1]_ is used to retrieve them (valid
                                         if your databank is HITRAN data). HAPI
                                         is embedded into RADIS. Check the version.

        parfuncfmt:                      # 'cdsd'
                                         # format to read tabulated partition function
                                         # file. If `USE_HAPI` is given as `parfunc`
                                         # parameter then this line should not be used.

        levels_iso1                      # path to energy levels (needed for non-eq)
                                         # calculations.
        levels_iso2                      # etc
        levels_iso4                      # etc

        levelsfmt                        # 'cdsd'
                                         # how to read the previous file.

    References
    ----------

    .. [1] `HAPI: The HITRAN Application Programming Interface <http://hitran.org/hapi>`_


    """

    config = getConfig()

    # Make sure it looks like a databank (path and format are given)
    try:
        config.get(dbname, "path")
        config.get(dbname, "format")
    except configparser.NoSectionError:
        msg = (
            "{1}\nDBFORMAT\n{0}\n".format(DBFORMAT, dbname)
            + "No databank named {0} in `{1}`. ".format(dbname, CONFIG_PATH)
            + "Available databanks: {0}. ".format(getDatabankList())
            + "See databank format above. More information in "
            + "https://radis.readthedocs.io/en/latest/lbl/lbl.html#configuration-file"
        )
        raise DatabankNotFound(msg)

    entries = dict(config.items(dbname))

    # Parse paths correctly
    entries["path"] = entries["path"].strip("\n").split("\n")

    # Merge all isotope-dependant levels into one dict
    iso_list = [k for k in entries if k.startswith("levels_iso")]
    if len(iso_list) > 0:
        levels = {}
        for k in iso_list:
            iso = float(k[10:])
            levels[iso] = entries.pop(k)
        entries["levels"] = levels
    else:
        if "levels" in entries:
            raise SyntaxError(
                "in {0}: `levels` replaced with ".format(CONFIG_PATH)
                + "`levels_iso#` where # is the isotope number"
            )

    return entries


def getDatabankList():
    """ Get all databanks available in ~/.radis"""

    config = getConfig()

    # Get databank path and format
    validdb = []
    for dbname in config.sections():
        try:
            config.get(dbname, "path")
            config.get(dbname, "format")
        except configparser.NoSectionError:
            # not a db
            continue
        except configparser.NoOptionError:
            # not a db
            continue
        # looks like a db. add to db
        validdb.append(dbname)
    return validdb


def addDatabankEntries(dbname, dict_entries, verbose=True):
    """Add database dbname with entries from dict_entries. If database
    already exists in ~/.radis, raises an error
    """

    # Get ~/.radis if exists, else create it
    try:
        dbnames = getDatabankList()
    except FileNotFoundError:
        # generate ~/.radis:
        dbnames = []
        open(CONFIG_PATH, "a").close()
        if verbose:
            print("Created ~/.radis in {0}".format(dirname(CONFIG_PATH)))

    # Check database doesnt exist
    if dbname in dbnames:
        raise ValueError(
            "Database already exists: {0}".format(dbname) + ". Cant add it"
        )

    # Add entries to parser
    config = configparser.ConfigParser()
    config[dbname] = {}

    if "info" in dict_entries:  # optional
        config[dbname]["info"] = dict_entries.pop("info")

    # ... Parse paths correctly
    if isinstance(dict_entries["path"], str):
        config[dbname]["path"] = dict_entries.pop("path")
    else:  # list
        config[dbname]["path"] = "\n       ".join(dict_entries.pop("path"))

    config[dbname]["format"] = dict_entries.pop("format")
    config[dbname]["parfuncfmt"] = dict_entries.pop("parfuncfmt")

    # Optional:
    # ... partition functions:
    if "parfunc" in dict_entries:
        config[dbname]["parfunc"] = dict_entries.pop("parfunc")

    # ... Split all isotopes in separate keys
    levels_dict = dict_entries.pop("levels", {})
    for iso, levels_iso in levels_dict.items():
        config[dbname]["levels_iso{0}".format(iso)] = levels_iso

    if "levelsfmt" in dict_entries:
        config[dbname]["levelsfmt"] = dict_entries.pop("levelsfmt")

    # Check nothing is left
    if dict_entries != {}:
        raise ValueError("Unexpected keys: {0}".format(list(dict_entries.keys())))

    # Write to ~/.radis
    # ... Note: what if there is a PermissionError here? Try/except pass?
    with open(CONFIG_PATH, "a") as configfile:
        configfile.write("\n")
        config.write(configfile)
    if verbose:
        print("Added {0} database in {1}".format(dbname, CONFIG_PATH))

    return


def diffDatabankEntries(dict_entries1, dict_entries2, verbose=True):
    """Compare two Databank entries under dict format (i.e: output of
    getDatabankEntries)

    Returns None if no differences are found, or the first different key
    """

    k = None

    try:
        verbose_compare = "if_different" if verbose else False
        assert len(dict_entries1) == len(dict_entries2)
        assert (
            compare_lists(
                list(dict_entries1.keys()),
                list(dict_entries2.keys()),
                verbose=verbose_compare,
            )
            == 1
        )
        for k in dict_entries1.keys():
            v1 = dict_entries1[k]
            v2 = dict_entries2[k]
            if k in ["info", "format", "parfunc", "parfuncfmt", "levelsfmt"]:
                assert v1.lower() == v2.lower()
            elif k in ["path"]:
                assert (
                    compare_lists(
                        [stdpath(path1).lower() for path1 in v1],
                        [stdpath(path2).lower() for path2 in v2],
                        verbose=verbose_compare,
                    )
                    == 1
                )
            elif k in ["levels"]:
                assert (
                    compare_dict(
                        v1,
                        v2,
                        compare_as_paths=list(v1.keys()),
                        verbose=verbose_compare,
                    )
                    == 1
                )
            else:
                raise ValueError("Unexpected key:", k)

        return None

    except AssertionError:
        if verbose:
            print("Key doesnt match:", k)
        return k


def printDatabankEntries(dbname, crop=200):
    """Print databank info


    Parameters
    ----------

    dbname: str
        database name in ~/.radis

    crop: int
        if > 0, cutoff entries larger than that

    """
    entries = getDatabankEntries(dbname)
    print(dbname, "\n-------")
    for k, v in entries.items():
        # Add extra arguments
        args = []
        if k == "levelszpe":
            args.append("cm-1")
        v = "{0}".format(v)
        if len(v) > crop and crop > 0:
            v = v[:crop] + "..."
        # Print item
        print(k, ":", v, *args)


def printDatabankList():
    """ Print all databanks available in ~/.radis """
    try:
        print("Databanks in {0}: ".format(CONFIG_PATH), ",".join(getDatabankList()))
        for dbname in getDatabankList():
            print("\n")
            printDatabankEntries(dbname)
    except FileNotFoundError:
        print("No config file {0}".format(CONFIG_PATH))
        # it's okay


# %% Test

if __name__ == "__main__":
    from radis.test.misc.test_config import _run_testcases

    _run_testcases(verbose=True)
